import 'src/sudokuer_type.dart';

SudokuList get _initSudokuList => SudokuList.filled(9, List<int>.filled(9, 0));

TagsLine get _initTagsLine => TagsLine.filled(9, List<bool>.filled(9, false));

get _initTagsBlock => TagsBlock.filled(
    3, List<List<bool>>.filled(3, List<bool>.filled(9, false)));

SudokuList toSudokuList(Board list) {
  final _list = _initSudokuList;
  for (var i = 0; i < 9; i++) {
    for (var j = 0; j < 9; j++) {
      var digit = list[i][j];
      final columnList = _list[i].toList();
      if (digit is String && int.tryParse(digit) != null) {
        digit = int.parse(digit);
      }

      if (digit is int) {
        columnList[j] = digit;
        _list[i] = columnList;
      }
    }
  }
  return _list;
}

List _initTags(SudokuList list) {
  final _tagsRow = _initTagsLine;
  final _tagsCol = _initTagsLine;
  final _tagsBlock = _initTagsBlock;
  final List<List<int>> _spaces = [];
  for (var i = 0; i < 9; i++) {
    for (var j = 0; j < 9; j++) {
      final digit = list[i][j];
      if (digit == 0) {
        _spaces.add([i, j]);
      } else {
        _setTags(i, j, digit, _tagsRow, _tagsCol, _tagsBlock);
      }
    }
  }
  return [_tagsRow, _tagsCol, _tagsBlock, _spaces];
}

void _setTags(int i, int j, int digit, tagsRow, tagsCol, tagsBlock,
    {markStatus = true}) {
  _setTagsRow(i, digit, tagsRow, markStatus: markStatus);
  _setTagsCol(j, digit, tagsCol, markStatus: markStatus);
  _setTagsBlock(i, j, digit, tagsBlock, markStatus: markStatus);
}

void _setTagsRow(int i, int digit, TagsLine tagsRow, {markStatus = true}) {
  final tagsRowItem = [...tagsRow[i]];
  tagsRowItem[digit - 1] = markStatus;
  tagsRow[i] = tagsRowItem;
}

void _setTagsCol(int j, int digit, List<List<bool>> tagsCol,
    {markStatus = true}) {
  final tagsColItem = [...tagsCol[j]];
  tagsColItem[digit - 1] = markStatus;
  tagsCol[j] = tagsColItem;
}

int _getBlockX(x) {
  return (x / 3).floor();
}

int _getBlockY(y) {
  return (y / 3).floor();
}

void _setTagsBlock(int x, int y, int digit, TagsBlock tagsBlock,
    {markStatus = true}) {
  final blockX = _getBlockX(x);
  final blockY = _getBlockY(y);
  final tagsBlockYItem = [...tagsBlock[blockX]];
  final tagsBlockItem = [...tagsBlockYItem[blockY]];
  tagsBlockItem[digit - 1] = markStatus;
  tagsBlockYItem[blockY] = tagsBlockItem;
  tagsBlock[blockX] = tagsBlockYItem;
}

setSudokuList(int i, int j, int digit, SudokuList list) {
  final colList = [...list[i]];
  colList[j] = digit;
  list[i] = colList;
}

/// Sudoku solver
void solver(SudokuList list) {
  final _initTagsList = _initTags(list);
  final _tagsRow = _initTagsList[0];
  final _tagsCol = _initTagsList[1];
  final _tagsBlock = _initTagsList[2];
  final _spaces = _initTagsList[3];
  var _valid = false;

  bool _dfs(int pos) {
    if (pos == _spaces.length) {
      _valid = true;
      return false;
    }

    var space = _spaces[pos];
    final x = space[0];
    final y = space[1];
    final blockX = _getBlockX(x);
    final blockY = _getBlockY(y);
    for (var digitIndex = 0; digitIndex < 9 && !_valid; ++digitIndex) {
      if (!_tagsRow[x][digitIndex] &&
          !_tagsCol[y][digitIndex] &&
          !_tagsBlock[blockX][blockY][digitIndex]) {
        final digit = digitIndex + 1;
        _setTags(x, y, digit, _tagsRow, _tagsCol, _tagsBlock);
        setSudokuList(x, y, digit, list);
        _dfs(pos + 1);
        _setTags(x, y, digit, _tagsRow, _tagsCol, _tagsBlock,
            markStatus: false);
      }
    }
    return true;
  }

  _dfs(0);
}

const difficultyLevelMap = {
  0: 0,
  1: 18,
  2: 27,
  3: 36,
  4: 54,
};

int _randomIndexGenerator() {
  var numberList = <int>[0, 1, 2, 3, 4, 5, 6, 7, 8]..shuffle();
  return numberList[0];
}

void _deduce(list) {
  final x = _randomIndexGenerator();
  final y = _randomIndexGenerator();
  if (list[x][y] != 0) {
    setSudokuList(x, y, 0, list);
  } else {
    _deduce(list);
  }
}

/// Sudoku generator
///
/// difficultyLevel: Solution Grid - 0, Beginner - 1, Easy - 2, Medium - 3, Hard - 4
SudokuList generator({int difficultyLevel = 2}) {
  final _list = _initSudokuList;
  final _emptyCell = difficultyLevelMap.containsKey(difficultyLevel)
      ? difficultyLevelMap[difficultyLevel]
      : difficultyLevelMap[2];

  _list[0] = [1, 2, 3, 4, 5, 6, 7, 8, 9]..shuffle();

  solver(_list);

  for (var i = 0; i < _emptyCell!; i++) {
    _deduce(_list);
  }

  return _list;
}

class Sudokuer {
  late SudokuList _sudokuList;

  get sudokuList => _sudokuList;

  Sudokuer({Board? list, difficultyLevel = 2}) {
    if (list == null) {
      _sudokuList = generator(difficultyLevel: difficultyLevel);
    } else {
      _sudokuList = list is SudokuList ? list : toSudokuList(list);
    }
  }
}
